@@ -1,5 +1,5 @@
 /*
-   (C)Copyright 2012 Simon Urbanek.
+   (C)Copyright 2012-2019 Simon Urbanek.
 
    Released under GPL v2, no warranties.
 
@@ -45,6 +45,12 @@ static int wsock_up = 0;
 #endif
 
 #ifdef USE_TLS
+
+/* OpenSSL 3.x requires this */
+#ifndef OPENSSL_SUPPRESS_DEPRECATED
+#define OPENSSL_SUPPRESS_DEPRECATED 1
+#endif
+
 #include <openssl/rsa.h>
 #include <openssl/rand.h>
 #include <openssl/ssl.h>
@@ -123,19 +129,53 @@ static int sock_recv(rsconn_t *c, void *buf, int len) {
 }
 
 #ifdef USE_TLS
+static int rsc_abort(rsconn_t *c, const char *reason);
+
 static int tls_send(rsconn_t *c, const void *buf, int len) {
-    return SSL_write((SSL*)c->tls, buf, len);
+    if (c->intr)
+	rsc_abort(c, "previous operation was interrupted, connection aborted");
+
+    /* SSL can fail with SSL_ERROR_WANT_READ/WRITE which is retriable */
+    while (1) {
+        int n = SSL_write((SSL*)c->tls, buf, len);
+        if (n <= 0) {
+            int serr = SSL_get_error((SSL*)c->tls, n);
+            if (serr != SSL_ERROR_WANT_READ && serr != SSL_ERROR_WANT_WRITE)
+                return n;
+            /* means we should allow interrupt as it can retry indefinitely */
+            c->intr = 1;
+            R_CheckUserInterrupt();
+            c->intr = 0;
+	} else return n;
+    }
+    return -1; /* unreachable */
 }
 
 static int tls_recv(rsconn_t *c, void *buf, int len) {
-    return SSL_read((SSL*)c->tls, buf, len);
+    if (c->intr)
+	rsc_abort(c, "previous operation was interrupted, connection aborted");
+
+    /* SSL can fail with SSL_ERROR_WANT_READ/WRITE which is retriable */
+    while (1) {
+	int n = SSL_read((SSL*)c->tls, buf, len);
+	if (n <= 0) {
+	    int serr = SSL_get_error((SSL*)c->tls, n);
+	    if (serr != SSL_ERROR_WANT_READ && serr != SSL_ERROR_WANT_WRITE)
+		return n;
+	    /* means we should allow interrupt as it can retry indefinitely */
+	    c->intr = 1;
+	    R_CheckUserInterrupt();
+	    c->intr = 0;
+	} else return n;
+    }
+    return -1; /* unreachable */
 }
 
 static int first_tls = 1;
 
 #include <openssl/err.h>
 
-static void init_tls() {
+static void init_tls(void) {
     if (first_tls) {
 	SSL_library_init();	
 	SSL_load_error_strings();
@@ -143,13 +183,26 @@ static void init_tls() {
     }
 }
 
-static int tls_upgrade(rsconn_t *c) {
+static int tls_upgrade(rsconn_t *c, int verify, const char *chain, const char *key, const char *ca) {
     SSL *ssl;
     SSL_CTX *ctx;
     if (first_tls)
 	init_tls();
     ctx = SSL_CTX_new(SSLv23_client_method());
     SSL_CTX_set_mode(ctx, SSL_MODE_AUTO_RETRY);
+    if (chain && SSL_CTX_use_certificate_chain_file(ctx, chain) != 1) {
+	Rf_warning("Cannot load certificate chain from file %s", chain);
+	return -1;
+    }
+    if (key && SSL_CTX_use_PrivateKey_file(ctx, key, SSL_FILETYPE_PEM) != 1) {
+	Rf_warning("Cannot load certificate key from file %s", key);
+	return -1;
+    }
+    if (ca && SSL_CTX_load_verify_locations(ctx, ca, 0) != 1) {
+	Rf_warning("Cannot load CA certificates from file %s", chain);
+	return -1;
+    }
+    SSL_CTX_set_verify(ctx, (verify == 0) ? SSL_VERIFY_NONE : SSL_VERIFY_PEER, 0);
     c->tls = ssl = SSL_new(ctx);
     c->send = tls_send;
     c->recv = tls_recv;
@@ -161,7 +214,7 @@ static int tls_upgrade(rsconn_t *c) {
 
 /* we split alloc and connect so alloc can be done on the main thread
    and connect on a separate one */
-static rsconn_t *rsc_alloc() {
+static rsconn_t *rsc_alloc(void) {
     rsconn_t *c = (rsconn_t*) calloc(sizeof(rsconn_t), 1);
 #ifdef WIN32
     if (!wsock_up) {
@@ -229,7 +282,7 @@ static rsconn_t *rsc_connect_ex(const char *host, int port, rsconn_t *c) {
 	struct timeval tv;
 	tv.tv_sec  = 0;
 	tv.tv_usec = 200000; /* 200ms */
-	setsockopt(c->s, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));
+	setsockopt(c->s, SOL_SOCKET, SO_RCVTIMEO, (const char *) &tv, sizeof(tv));
     }
 #endif
     if (c->s != -1 && !connected) {
@@ -401,6 +454,7 @@ static long rsc_read(rsconn_t *c, void *buf, long needed) {
     return (long) (ptr - (char*) buf);
 }
 
+/* Note: OC handshake also uses the slurp buffer as scratch */
 static char slurp_buffer[65536];
 
 static long rsc_slurp(rsconn_t *c, long needed) {
@@ -416,6 +470,7 @@ static long rsc_slurp(rsconn_t *c, long needed) {
 
 /* Rserve protocol */
 #include "RSprotocol.h"
+#include "qap.h"
 
 /* --- R API -- */
 
@@ -426,16 +481,19 @@ static void rsconn_fin(SEXP what) {
     if (c) rsc_close(c);
 }
 
-/* FIXME: hack for now -- most of this client works only on little-endian machines
-   but there is some feeble effort to fix that -- in the meantime it's just noop */
-#define itop(X) X
+static void setAttrib_(SEXP x, const char *sym, SEXP sVal) {
+    PROTECT(sVal);
+    Rf_setAttrib(x, Rf_install(sym), sVal);
+    UNPROTECT(1);
+}
 
-SEXP RS_connect(SEXP sHost, SEXP sPort, SEXP useTLS, SEXP sProxyTarget, SEXP sProxyWait) {
-    int port = asInteger(sPort), use_tls = (asInteger(useTLS) == 1), px_get_slot = (asInteger(sProxyWait) == 0);
+SEXP RS_connect(SEXP sHost, SEXP sPort, SEXP useTLS, SEXP sProxyTarget, SEXP sProxyWait, SEXP sVerify,
+		SEXP sChainFile, SEXP sKeyFile, SEXP sCAFile) {
+    int port = asInteger(sPort), use_tls = (asInteger(useTLS) == 1), px_get_slot = (asInteger(sProxyWait) == 0), n;
     const char *host;
     char idstr[32];
     rsconn_t *c;
-    SEXP res;
+    SEXP res, caps = R_NilValue;
 
     if (port < 0 || port > 65534)
 	Rf_error("Invalid port number");
@@ -460,9 +518,17 @@ SEXP RS_connect(SEXP sHost, SEXP sPort, SEXP useTLS, SEXP sProxyTarget, SEXP sPr
     if (!c)
 	Rf_error("cannot connect to %s:%d", host, port);
 #ifdef USE_TLS
-    if (use_tls && tls_upgrade(c) != 1) {
-	rsc_close(c);
-	Rf_error("TLS handshake failed");
+    if (use_tls) {
+	const char *chain = ((TYPEOF(sChainFile) == STRSXP) && LENGTH(sChainFile) > 0) ? CHAR(STRING_ELT(sChainFile, 0)) : 0;
+	const char *key = ((TYPEOF(sKeyFile) == STRSXP) && LENGTH(sKeyFile) > 0) ? CHAR(STRING_ELT(sKeyFile, 0)) : 0;
+	const char *ca = ((TYPEOF(sCAFile) == STRSXP) && LENGTH(sCAFile) > 0) ? CHAR(STRING_ELT(sCAFile, 0)) : 0;
+	if ((n = tls_upgrade(c, asInteger(sVerify), chain, key, ca)) != 1) {
+	    int serr = SSL_get_error((SSL*)c->tls, n);
+	    unsigned long err = ERR_get_error();
+	    const char *es = ERR_error_string(err, 0);
+	    rsc_close(c);
+	    Rf_error("TLS handshake failed (SSL_error=%d; %s)", serr, es);
+	}
     }
 #endif	
     if (rsc_read(c, idstr, 32) != 32) {
@@ -495,20 +561,54 @@ SEXP RS_connect(SEXP sHost, SEXP sPort, SEXP useTLS, SEXP sProxyTarget, SEXP sPr
 	    Rf_error("Handshake failed - ID string not received (after CMD_PROXY_TARGET)");
 	}
     }
-    if (memcmp(idstr, "Rsrv", 4) || memcmp(idstr + 8, "QAP1", 4)) {
-	rsc_close(c);
-	Rf_error("Handshake failed - unknown protocol");
-    }
+    /* OC mode */
+    if (((const int*)idstr)[0] == itop(CMD_OCinit)) {
+	int sb_len;
+	struct phdr *hdr = (struct phdr *) idstr;
+	hdr->len = itop(hdr->len);
+	if (hdr->res || hdr->dof || hdr->len > sizeof(slurp_buffer) || hdr->len < 16) {
+	    rsc_close(c);
+	    Rf_error("Handshake failed - invalid RsOC OCinit message");
+	}
+	sb_len = 32 - sizeof(struct phdr);
+	memcpy(slurp_buffer, idstr + sizeof(struct phdr), sb_len);
+	if (rsc_read(c, slurp_buffer + sb_len, hdr->len - sb_len) != hdr->len - sb_len) {
+	    rsc_close(c);
+	    Rf_error("Handshake failed - truncated RsOC OCinit message");
+	} else {
+	    unsigned int *ibuf = (unsigned int*) slurp_buffer;
+	    int par_type = PAR_TYPE(*ibuf);
+	    int is_large = (par_type & DT_LARGE) ? 1 : 0;
+	    if (is_large) par_type ^= DT_LARGE;
+	    if (par_type != DT_SEXP) {
+		rsc_close(c);
+		Rf_error("Handshake failed - invalid payload in OCinit message");
+	    }
+	    ibuf += is_large + 1;
+	    caps = QAP_decode(&ibuf);
+	    if (caps != R_NilValue)
+		PROTECT(caps);
+	}
+    } else {
+	if (memcmp(idstr, "Rsrv", 4) || memcmp(idstr + 8, "QAP1", 4)) {
+	    rsc_close(c);
+	    Rf_error("Handshake failed - unknown protocol");
+	}
 
-    /* supported range 0100 .. 0103 */
-    if (memcmp(idstr + 4, "0100", 4) < 0 || memcmp(idstr + 4, "0103", 4) > 0) {
-	rsc_close(c);
-	Rf_error("Handshake failed - server protocol version too high");
+	/* supported range 0100 .. 0103 */
+	if (memcmp(idstr + 4, "0100", 4) < 0 || memcmp(idstr + 4, "0103", 4) > 0) {
+	    rsc_close(c);
+	    Rf_error("Handshake failed - server protocol version too high");
+	}
     }
 
     res = PROTECT(R_MakeExternalPtr(c, R_NilValue, R_NilValue));
     setAttrib(res, R_ClassSymbol, mkString("RserveConnection"));
     R_RegisterCFinalizer(res, rsconn_fin);
+    if (caps != R_NilValue) {
+	setAttrib_(res, "capabilities", caps);
+	UNPROTECT(1);
+    }	  
     UNPROTECT(1);
     return res;
 }
@@ -557,8 +657,6 @@ static const char *rs_status_string(int status) {
     return "(unknown error code)";
 }
 
-#include "qap.h"
-
 #ifdef USE_THREADS
 
 /* threaded version - can be run ona separate threads, does not use
@@ -672,7 +770,6 @@ static long get_hdr(SEXP sc, rsconn_t *c, struct phdr *hdr) {
 	if (hdr->cmd & CMD_OOB) {
 	    SEXP res, ee = R_NilValue;
 	    unsigned int *ibuf;
-	    int upc = 0;
 	    PROTECT(res = allocVector(RAWSXP, tl));
 	    if (rsc_read(c, RAW(res), tl) != tl) {
 		c->in_cmd = 0;
@@ -682,22 +779,26 @@ static long get_hdr(SEXP sc, rsconn_t *c, struct phdr *hdr) {
 	    ibuf = (unsigned int*) RAW(res);
 	    /* FIXME: we assume that we get encoded SEXP - we should check ... */
 	    ibuf += 1;
-	    res = QAP_decode(&ibuf, &upc);
+	    res = QAP_decode(&ibuf);
+	    UNPROTECT(1); /* original RAW res */
+	    PROTECT(res); /* result */
 
 	    /* FIXME: Rserve has a bug(?) that sets CMD_RESP on OOB commands so we clear it for now ... */
 	    hdr->cmd &= ~CMD_RESP;
 
-	    if (IS_OOB_SEND(hdr->cmd) && c->oob_send_cb)
+	    if (IS_OOB_SEND(hdr->cmd) && c->oob_send_cb != R_NilValue)
 		PROTECT(ee = lang3(c->oob_send_cb, ScalarInteger(OOB_USR_CODE(hdr->cmd)), res));
-	    if (IS_OOB_MSG(hdr->cmd) && c->oob_msg_cb)
+	    if (IS_OOB_MSG(hdr->cmd) && c->oob_msg_cb != R_NilValue)
 		PROTECT(ee = lang3(c->oob_msg_cb, ScalarInteger(OOB_USR_CODE(hdr->cmd)), res));
+#ifdef RC_DEBUG
 	    Rprintf(" - OOB %x %s (%d) %d\n", hdr->cmd, IS_OOB_SEND(hdr->cmd) ? "send" : "other", OOB_USR_CODE(hdr->cmd), (int) tl);
-	    if (ee != R_NilValue) {
+#endif
+	    if (ee != R_NilValue) { /* OOB send or msg - we ignore anything else */
 		res = eval(ee, R_GlobalEnv);
 		if (IS_OOB_MSG(hdr->cmd)) {
 		    struct phdr rhdr;
 		    long pl = QAP_getStorageSize(res);
-		    SEXP outv = allocVector(RAWSXP, pl);
+		    SEXP outv = PROTECT(allocVector(RAWSXP, pl));
 		    int isx = pl > 0x7fffff;
 		    unsigned int *oh = (unsigned int*) RAW(outv);
 		    unsigned int *ot = QAP_storeSEXP(oh + (isx ? 2 : 1), res, pl);
@@ -715,10 +816,11 @@ static long get_hdr(SEXP sc, rsconn_t *c, struct phdr *hdr) {
 		    rsc_write(c, &rhdr, sizeof(rhdr));
 		    if (pl) rsc_write(c, RAW(outv), pl);
 		    rsc_flush(c);
+		    UNPROTECT(1); /* outv */
 		}
-		UNPROTECT(1);
+		UNPROTECT(1); /* ee */
 	    }
-	    UNPROTECT(upc + 1);
+	    UNPROTECT(1); /* res */
 	    continue;
 	}
 	break;
@@ -731,6 +833,68 @@ static long get_hdr(SEXP sc, rsconn_t *c, struct phdr *hdr) {
     return tl;
 }
 
+SEXP RS_eval_qap(SEXP sc, SEXP what, SEXP sWait) {
+    SEXP res = R_NilValue;
+    rsconn_t *c;
+    int async = (asInteger(sWait) == 0);
+
+    if (!inherits(sc, "RserveConnection")) Rf_error("invalid connection");
+    c = (rsconn_t*) EXTPTR_PTR(sc);
+    if (!c) Rf_error("invalid NULL connection");
+    if (!async && c->in_cmd) Rf_error("uncollected result from previous command, remove first");
+
+    {
+	struct phdr rhdr;
+	long pl   = QAP_getStorageSize(what), tl;
+	SEXP outv = PROTECT(allocVector(RAWSXP, pl));
+	int isx   = pl > 0x7fffff;
+	unsigned int *oh = (unsigned int*) RAW(outv);
+	unsigned int *ot = QAP_storeSEXP(oh + (isx ? 2 : 1), what, pl);
+
+	pl = sizeof(int) * (long) (ot - oh);
+	rhdr.cmd = CMD_eval;
+	/* If the call is OCref then it's OCcall and not eval ... */
+	if (TYPEOF(what) == LANGSXP && inherits(CAR(what), "OCref")) rhdr.cmd = CMD_OCcall;
+	rhdr.len = pl;
+	rhdr.dof = 0;
+#ifdef __LP64__
+	rhdr.res = pl >> 32;
+#else
+	rhdr.res = 0;
+#endif
+	oh[0] = SET_PAR(DT_SEXP | (isx ? DT_LARGE : 0), pl - (isx ? 8 : 4));
+	if (isx) oh[1] = (pl - 8) >> 24;
+	rsc_write(c, &rhdr, sizeof(rhdr));
+	if (pl) rsc_write(c, RAW(outv), pl);
+	rsc_flush(c);
+	UNPROTECT(1); /* outv */
+	outv = 0;
+
+	if (async) {
+	    c->in_cmd++;
+	    return R_NilValue;
+	}
+	tl = get_hdr(sc, c, &rhdr);
+	res = PROTECT(allocVector(RAWSXP, tl));
+	if (rsc_read(c, RAW(res), tl) != tl) {
+	    RS_close(sc);
+	    Rf_error("read error reading payload of the eval result");
+	} else {
+	    unsigned int *ibuf = (unsigned int*) RAW(res);
+	    int par_type = PAR_TYPE(*ibuf);
+	    int is_large = (par_type & DT_LARGE) ? 1 : 0;
+	    if (is_large) par_type ^= DT_LARGE;
+	    if (par_type != DT_SEXP)
+		Rf_error("invalid result type coming from eval");
+	    ibuf += is_large + 1;
+	    res = QAP_decode(&ibuf);
+	}
+	UNPROTECT(1);
+    }
+
+    return res;
+}
+
 SEXP RS_eval(SEXP sc, SEXP what, SEXP sWait) {
     SEXP res;
     rsconn_t *c;
@@ -755,11 +919,12 @@ SEXP RS_eval(SEXP sc, SEXP what, SEXP sWait) {
 	return R_NilValue;
     }
     tl = get_hdr(sc, c, &hdr);
-    res = allocVector(RAWSXP, tl);
+    res = PROTECT(allocVector(RAWSXP, tl));
     if (rsc_read(c, RAW(res), tl) != tl) {
 	RS_close(sc);
 	Rf_error("read error reading payload of the eval result");
     }
+    UNPROTECT(1);
     return res;
 }
 
@@ -819,8 +984,8 @@ SEXP RS_collect(SEXP sc, SEXP s_timeout) {
 	/* both sc and c are set to the node and the structure */
 	tl = get_hdr(sc, c, &hdr);
 	res = PROTECT(allocVector(RAWSXP, tl));
-	setAttrib(res, install("rsc"), sc);
-	if (rdy >= 0) setAttrib(res, install("index"), ScalarInteger(rdy + 1));
+	setAttrib_(res, "rsc", sc);
+	if (rdy >= 0) setAttrib_(res, "index", ScalarInteger(rdy + 1));
 	if (rsc_read(c, RAW(res), tl) != tl) {
 	    RS_close(sc);
 	    Rf_error("read error reading payload of the eval result");
@@ -830,6 +995,18 @@ SEXP RS_collect(SEXP sc, SEXP s_timeout) {
     }
 }
 
+SEXP RS_decode(SEXP sWhat) {
+    unsigned int *ibuf = (unsigned int*) RAW(sWhat);
+    int par_type = PAR_TYPE(*ibuf);
+    int is_large = (par_type & DT_LARGE) ? 1 : 0;
+    if (is_large) par_type ^= DT_LARGE;
+    if (par_type != DT_SEXP)
+	Rf_error("invalid result - must be DT_SEXP");
+    ibuf += is_large + 1;
+    /* FIXME: we don't check if we message is complete */
+    return QAP_decode(&ibuf);
+}
+
 SEXP RS_assign(SEXP sc, SEXP what, SEXP sWait) {
     SEXP res;
     rsconn_t *c;
@@ -854,11 +1031,12 @@ SEXP RS_assign(SEXP sc, SEXP what, SEXP sWait) {
 	return R_NilValue;
     }
     tl = get_hdr(sc, c, &hdr);
-    res = allocVector(RAWSXP, tl);
+    res = PROTECT(allocVector(RAWSXP, tl));
     if (rsc_read(c, RAW(res), tl) != tl) {
 	RS_close(sc);
 	Rf_error("read error reading payload of the eval result");
     }
+    UNPROTECT(1);
     return res;
 }
 
@@ -907,13 +1085,17 @@ SEXP RS_ctrl_str(SEXP sc, SEXP sCmd, SEXP sPayload) {
     return ScalarLogical(TRUE);
 }
 
-SEXP RS_switch(SEXP sc, SEXP prot) {
+SEXP RS_switch(SEXP sc, SEXP prot, SEXP sVerify, SEXP sChainFile, SEXP sKeyFile, SEXP sCAFile) {
     rsconn_t *c;
 
     if (!inherits(sc, "RserveConnection")) Rf_error("invalid connection");
     c = (rsconn_t*) EXTPTR_PTR(sc);
     if (!c) Rf_error("invalid NULL connection");
     if (c->in_cmd) Rf_error("uncollected result from previous command, remove first");
+    const char *chain = ((TYPEOF(sChainFile) == STRSXP) && LENGTH(sChainFile) > 0) ? CHAR(STRING_ELT(sChainFile, 0)) : 0;
+    const char *key = ((TYPEOF(sKeyFile) == STRSXP) && LENGTH(sKeyFile) > 0) ? CHAR(STRING_ELT(sKeyFile, 0)) : 0
+;
+    const char *ca = ((TYPEOF(sCAFile) == STRSXP) && LENGTH(sCAFile) > 0) ? CHAR(STRING_ELT(sCAFile, 0)) : 0;
     if (TYPEOF(prot) != STRSXP || LENGTH(prot) != 1)
 	Rf_error("invalid protocol specification");
 #ifdef USE_TLS
@@ -933,7 +1115,7 @@ SEXP RS_switch(SEXP sc, SEXP prot) {
 	tl = get_hdr(sc, c, &hdr);
 	if (tl)
 	    rsc_slurp(c, tl);
-	if (tls_upgrade(c) != 1) {
+	if (tls_upgrade(c, asInteger(sVerify), chain, key, ca) != 1) {
 	    RS_close(sc);
 	    Rf_error("TLS negotitation failed");
 	}
@@ -971,11 +1153,12 @@ SEXP RS_authkey(SEXP sc, SEXP type) {
     rsc_write(c, key_type, strlen(key_type) + 1);
     rsc_flush(c);
     tl = get_hdr(sc, c, &hdr);
-    res = allocVector(RAWSXP, tl);
+    res = PROTECT(allocVector(RAWSXP, tl));
     if (rsc_read(c, RAW(res), tl) != tl ) {
 	RS_close(sc);
 	Rf_error("read error loading key payload");
     }
+    UNPROTECT(1);
     return res;
 }
 
@@ -1096,8 +1279,8 @@ SEXP RS_oob_cb(SEXP sc, SEXP send_cb, SEXP msg_cb, SEXP query) {
 	}
     }
     PROTECT(res = Rf_mkNamed(VECSXP, (const char *[]) { "send", "msg", "" }));
-    SET_VECTOR_ELT(res, 0, send_cb);
-    SET_VECTOR_ELT(res, 1, msg_cb);
+    SET_VECTOR_ELT(res, 0, c->oob_send_cb);
+    SET_VECTOR_ELT(res, 1, c->oob_msg_cb);
     UNPROTECT(1);
     return res;    
 }
@@ -1114,9 +1297,9 @@ SEXP RS_print(SEXP sc) {
     if (!c)
 	Rprintf(" <NULL> **invalid** RserveConnection\n");
     else if (c->s == -1)
-	Rprintf(" Closed Rserve connection %p\n", c);
+	Rprintf(" Closed Rserve connection %p\n", (void*) c);
     else
-	Rprintf(" Rserve %s connection %p (socket %d, queue length %d)\n", c->tls ? "TLS/QAP1" : "QAP1", c, c->s, c->in_cmd);
+	Rprintf(" Rserve %s connection %p (socket %d, queue length %d)\n", c->tls ? "TLS/QAP1" : "QAP1", (void*) c, (int) c->s, (int) c->in_cmd);
     return sc;
 }
 
